/* noreusal.c, v1.0, 15.11.1993
 *
 * This program is copyright Tomasz Surmacz 1993 (tsurmacz@ict.pwr.wroc.pl).
 * The author disclaims all responsibility or liability with respect to it's
 * usage or its effect upon hardware or computer systems, and maintains
 * copyright on this source.
 *
 * Permission is granted to copy and use this source in unmodified version
 * as part of the CRACKLIB package by Alec Muffet or any other package
 * improving password security on your system.
 */

static char vers_id[] = "noreusal.c : v1.0 Tomasz Surmacz 15 Nov 1993";

#include <stdio.h>
#include <pwd.h>
#include <string.h>
#include "language.h"

#include <sys/types.h>
#include <sys/stat.h>

#ifndef PASSWD_HISTORY_FILE
#define PASSWD_HISTORY_FILE "/etc/passwd.hist"
#endif

/*
 *	fgetSpwent() - read simple/incomplete passwd(5) entry from a file.
 *      Only 'user:passwd:' part is required, the rest is ignored. (It may be
 *	used for comments).
 *      user field may be a real user or '*' in which case it will match any
 *	user when making comparisons.
 *	Standalone comments can be put in this file in form of '#:*: Comment'
 */
struct passwd *
fgetSpwent(f)
FILE	*f;		/* Pointer to open simplepasswd format file */
{
	static struct passwd pwdata;	/* Return data */
	static char ibuf[BUFSIZ];	/* Input and return data buffer */
	char *p;

	bzero((char *)&pwdata, sizeof(pwdata));
	pwdata.pw_name = pwdata.pw_passwd = pwdata.pw_comment =
	pwdata.pw_gecos = pwdata.pw_dir = pwdata.pw_shell = "";

	if (fgets(ibuf, sizeof(ibuf), f) == NULL)
		return 0;
	if ((p = strchr(ibuf, '\n')) == NULL) {	/* Zap newline */
		fprintf(stderr,"Ill-formed old-passwd entry \"%s\".\n", ibuf);
		return 0;
	}
	else
		*p = 0;
#define	skipc while (*p && *p != ':' && *p != '\n') ++p; if (*p) *p++ = 0
	p = ibuf;
	pwdata.pw_name = p;	skipc;
	pwdata.pw_passwd = p;	skipc;
	return(&pwdata);
#undef	skipc
}

int HTry(username, password)
    register char *username;
    register char *password;
{
    FILE *f;
    struct passwd *pwdata;
    int m;

    if ((f=fopen(PASSWD_HISTORY_FILE,"rt"))==NULL) {
	fprintf(stderr,ERR_NOHISTFILE);
	return 0;  /* print the warning and accept the password */
    }

    while (pwdata=fgetSpwent(f)) {

	/* first - do not compare with null passwords (like 'sync::...' */
	if (pwdata->pw_passwd) {
		m=1;
		/* if the username is '*' or matches the given user - check if
		 * the password matches. If yes - return 1 in case of
		 * 'system-wide' match or 2 - if user matches too.
		 */
		if (pwdata->pw_name[0]=='*' || (m=2,!strcmp(pwdata->pw_name,username))) {
			if (!strcmp(pwdata->pw_passwd, crypt(password, pwdata->pw_passwd))) {
				fclose (f);
				return m;
			}
		}
	}
    }
    fclose(f);
    return 0;
}


char *
FascistHistory(password, uid)
    char *password;
    int uid;
{
    int i;
    int j;
    int wc;
    struct passwd *pwp;

    if (!(pwp = getpwuid(uid))) {
	return (MSG_PWNOTREG);
    }

    if (HTry(pwp->pw_name, password)) {
	return (MSG_SAMEAGAIN);
    }

    return (char*) 0;
}

/* HistUpdate() - library function to implement full 'no password reusal'
 * policy.  It should be called from the password changing program just after
 * the password has been changed.  It adds given (encrypted) password to the
 * 'old passwords' database so it cannot be used again.
 */

int HistUpdate(username, cryptedpassword)
    register char *username;
    register char *cryptedpassword;
{
    FILE *f;
    struct passwd *pwdata;
    int m;
    struct stat statbuf;

    /* update only if file has write permission for owner.  (passwd changing 
     * program runs as suid to root, 'access()' would check if the user has
     * write permission which would be 'no', so we must check stat() instead)
     */

    stat(PASSWD_HISTORY_FILE,&statbuf);
    if (!(statbuf.st_mode & S_IWUSR))
	return 1;	

#if 0
    if (access(PASSWD_HISTORY_FILE,2))
	return 1;
#endif

    if ((f=fopen(PASSWD_HISTORY_FILE,"at"))==NULL) {
	fprintf(stderr,ERR_NOHISTUPDATE);
	return 2;  /* print warning and exit */
    }

    fprintf(f,"%s:%s:\n",username,cryptedpassword);
    fclose (f);
    return 0;
}
